@@ -63,7 +63,7 @@ class BusConnection (rsb.eventprocessing.BroadcastProcessor):
 
     def __init__(self,
                  host = None, port = None, socket_ = None,
-                 isServer = False):
+                 isServer = False, tcpnodelay = True):
         """
         @param host: Hostname or address of the bus server.
         @type host: str or None
@@ -71,6 +71,11 @@ class BusConnection (rsb.eventprocessing.BroadcastProcessor):
         @type port: int or None
         @param socket_: A socket object through which the new
                         connection should access the bus.
+        @param isServer: if True, the created object will perform the server
+                         part of the handshake protocol.
+        @type isServer: bool
+        @param tcpnodelay: If True, the socket will be set to TCP_NODELAY.
+        @type tcpnodelay: bool
         @see: L{getBusClientFor}, L{getBusServerFor}.
         """
         super(BusConnection, self).__init__()
@@ -84,6 +89,9 @@ class BusConnection (rsb.eventprocessing.BroadcastProcessor):
         self.__errorHook = None
 
         self.__active = False
+        self.__activeShutdown = False
+        
+        self.__lock = threading.RLock()
 
         # Create a socket connection or store the provided connection.
         if not host is None and not port is None:
@@ -95,6 +103,10 @@ class BusConnection (rsb.eventprocessing.BroadcastProcessor):
             self.__socket = socket_
         else:
             raise ValueError, 'specify either host and port or socket_'
+        if tcpnodelay:
+            self.__socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+        else:
+            self.__socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 0)
         self.__file = self.__socket.makefile()
 
         # Perform the client or server part of the handshake.
@@ -122,13 +134,16 @@ class BusConnection (rsb.eventprocessing.BroadcastProcessor):
 
     def receiveNotification(self):
         size = self.__file.read(size = 4)
+        if len(size) == 0:
+            self.__logger.info("Received EOF")
+            raise EOFError()
         if not (len(size) == 4):
-            raise RuntimeError, 'short read when receiving notification size'
+            raise RuntimeError, 'short read when receiving notification size (size: %s)' % len(size)
         size = ord(size[0])      \
             | ord(size[1]) << 8  \
             | ord(size[2]) << 16 \
             | ord(size[3]) << 24
-        self.__logger.info('Receiving notification of size %d' % size)
+        self.__logger.debug('Receiving notification of size %d', size)
         notification = self.__file.read(size = size)
         if not (len(notification) == size):
             raise RuntimeError, 'short read when receiving notification payload'
@@ -145,29 +160,36 @@ class BusConnection (rsb.eventprocessing.BroadcastProcessor):
         self.dispatch(notification)
 
     def receiveNotifications(self):
-        while self.__socket:
-            self.__logger.info('Receiving notifications')
+        while True:
+            self.__logger.debug('Receiving notifications')
             try:
                 self.doOneNotification()
+            except EOFError, e:
+                self.__logger.info("Received EOF while reading")
+                if not self.__activeShutdown:
+                    self.shutdown()
+                break
             except Exception, e:
-                # __socket is None => shutdown in progress
-                if self.__socket:
-                    self.__logger.warn('Receive error: %s', e)
-                    if not self.errorHook is None:
-                        self.errorHook(e)
+                self.__logger.warn('Receive error: %s', e)
+                break
+
+        if not self.errorHook is None:
+            self.errorHook(e)
+        return
 
     # sending
 
     def sendNotification(self, notification):
         size = len(notification)
-        self.__logger.info('Sending notification of size %d' % size)
+        self.__logger.info('Sending notification of size %d', size)
         size = ''.join((chr(size & 0x000000ff),
                         chr((size & 0x0000ff00) >> 8),
                         chr((size & 0x00ff0000) >> 16),
                         chr((size & 0xff000000) >> 24)))
-        self.__file.write(size)
-        self.__file.write(notification)
-        self.__file.flush()
+        with self.__lock:
+            self.__file.write(size)
+            self.__file.write(notification)
+            self.__file.flush()
 
     def notificationToBuffer(self, notification):
         return notification.SerializeToString()
@@ -182,39 +204,40 @@ class BusConnection (rsb.eventprocessing.BroadcastProcessor):
         if self.__active:
             raise RuntimeError, 'trying to activate active connection'
 
-        self.__thread = threading.Thread(target = self.receiveNotifications)
-        self.__thread.start()
+        with self.__lock:
 
-        self.__active = True
+            self.__thread = threading.Thread(target = self.receiveNotifications)
+            self.__thread.start()
+    
+            self.__active = True
+
+    def shutdown(self):
+        with self.__lock:
+            self.__activeShutdown = True
+            self.__socket.shutdown(socket.SHUT_WR)
 
     def deactivate(self):
-        if not self.__active:
-            raise RuntimeError, 'trying to deactivate inactive connection'
 
-        self.__active = False
-
-        # If necessary, close the socket, this will cause an exception
-        # in the notification receiver thread (unless we run in the
-        # context that thread).
-        self.__logger.info('Closing socket')
-        try:
-            if not self.__file is None:
+        with self.__lock:
+        
+            if not self.__active:
+                raise RuntimeError, 'trying to deactivate inactive connection'
+    
+            self.__active = False
+    
+            # If necessary, close the socket, this will cause an exception
+            # in the notification receiver thread (unless we run in the
+            # context that thread).
+            self.__logger.info('Closing socket')
+            try:
                 self.__file.close()
-            if not self.__socket is None:
-                socket_ = self.__socket
-                self.__socket = None
-                socket_.shutdown(socket.SHUT_RDWR)
-                socket_.close()
-        except Exception, e:
-            self.__logger.warn('Failed to close socket: %s', e)
+                self.__socket.close()
+            except Exception, e:
+                self.__logger.warn('Failed to close socket: %s', e)
 
-        # If there is a receiver thread and we are not running in its
-        # context, the thread should encounter an exception and exit
-        # eventually. We wait for that to happen.
-        if not self.__thread is None \
-                and not self.__thread is threading.currentThread():
-            self.__logger.info('Joining thread')
-            self.__thread.join()
+    def waitForDeactivation(self):
+        self.__logger.info('Joining thread')
+        self.__thread.join()
 
 class Bus (object):
     """
@@ -271,7 +294,13 @@ class Bus (object):
                 def __call__(_self, notification):
                     self.handleIncoming(( connection, notification ))
             connection.addHandler(Handler())
-            connection.errorHook = lambda exception: self.removeConnection(connection)
+            def removeAndDeactivate(exception):
+                self.removeConnection(connection)
+                try:
+                    connection.deactivate()
+                except Exception, e:
+                    self.__logger.warning("Error while deactivating connection %s: %s", connection, e)
+            connection.errorHook = removeAndDeactivate
             connection.activate()
 
     def removeConnection(self, connection):
@@ -283,13 +312,11 @@ class Bus (object):
         """
         self.__logger.info('Removing connection %s', connection)
 
-        try:
-            connection.deactivate()
-        finally:
-            with self.lock:
-                connection.removeHandler([ h for h in connection.handlers
-                                           if h.bus is self ][0])
+        with self.lock:
+            if connection in self.__connections:
                 self.__connections.remove(connection)
+                connection.removeHandler([ h for h in connection.handlers
+                                       if h.bus is self ][0])
 
     connections = property(getConnections)
 
@@ -349,10 +376,14 @@ class Bus (object):
 
             # Distribute the notification to remote participants via
             # network connections.
-            self._toConnections(notification)
+            failing = self._toConnections(notification)
             # Distribute the notification to participants in our own
             # process via InPushConnector instances.
             self._toConnectors(notification)
+        # there are only failing connection in case of an unorderly shutdown.
+        # So the shutdown protocol does not apply here and
+        # we can immediately call deactivate.
+        map(BusConnection.deactivate, failing)
 
     # State management
 
@@ -369,27 +400,38 @@ class Bus (object):
 
         with self.lock:
             self.__active = False
+            connectionsCopy = copy.copy(self.connections)
 
         # We do not have to lock the bus here, since
         # 1) removeConnection will do that for each connection
         # 2) the connection list will not be modified concurrently at
         #    this point
         self.__logger.info('Closing connections')
-        try:
-            map(self.removeConnection, copy.copy(self.connections))
-        except Exception, e:
-            self.__logger.error('Failed to close connections: %s', e)
+        for connection in connectionsCopy:
+            try:
+                self.removeConnection(connection)
+                connection.shutdown()
+                connection.waitForDeactivation()
+            except Exception, e:
+                self.__logger.error('Failed to close connections: %s', e)
 
     # Low-level helpers
 
     def _toConnections(self, notification, exclude = None):
+        failing = []
         for connection in self.connections:
             if not connection is exclude:
                 try:
                     connection.handle(notification)
                 except Exception, e:
-                    # TODO handle this
-                    self.__logger.warn('Failed to send to %s: %s', connection, e)
+                    self.__logger.warn('Failed to send to %s: %s; will close connection later',
+                                       connection, e)
+                    failing.append(connection)
+
+        # Removed connections for which sending the notification
+        # failed.
+        map(self.removeConnection, failing)
+        return failing
 
     def _toConnectors(self, notification):
         # Deliver NOTIFICATION to connectors which fulfill two
@@ -415,7 +457,7 @@ class Bus (object):
 __busClients = {}
 __busClientsLock = threading.Lock()
 
-def getBusClientFor(host, port, connector):
+def getBusClientFor(host, port, tcpnodelay, connector):
     """
     Return (creating it if necessary), a L{BusClient} for the endpoint
     designated by B{host} and B{port} and attach B{connector} to
@@ -428,14 +470,16 @@ def getBusClientFor(host, port, connector):
     @type host: str
     @param port: The port on which the bus server listens.
     @type port: int
+    @param tcpnodelay: If True, the socket will be set to TCP_NODELAY.
+    @type tcpnodelay: bool
     @param connector: A connector that should be attached to the bus
                       client.
     """
-    key = (host, port)
+    key = (host, port, tcpnodelay)
     with __busClientsLock:
         bus = __busClients.get(key)
         if bus is None:
-            bus = BusClient(host, port)
+            bus = BusClient(host, port, tcpnodelay)
             __busClients[key] = bus
             bus.activate()
             bus.addConnector(connector)
@@ -450,22 +494,24 @@ class BusClient (Bus):
 
     @author: jmoringe
     """
-    def __init__(self, host, port):
+    def __init__(self, host, port, tcpnodelay):
         """
         @param host: A hostname or address of the node on which the
                      bus server listens.
         @type host: str
         @param port: The port on which the new bus server listens.
         @type port: int
+        @param tcpnodelay: If True, the socket will be set to TCP_NODELAY.
+        @type tcpnodelay: bool
         """
         super(BusClient, self).__init__()
 
-        self.addConnection(BusConnection(host, port))
+        self.addConnection(BusConnection(host, port, tcpnodelay=tcpnodelay))
 
 __busServers = {}
 __busServersLock = threading.Lock()
 
-def getBusServerFor(host, port, connector):
+def getBusServerFor(host, port, tcpnodelay, connector):
     """
     Return (creating it if necessary), a L{BusServer} for the endpoint
     designated by B{host} and B{port} and attach B{connector} to
@@ -480,14 +526,16 @@ def getBusServerFor(host, port, connector):
     @param port: The port to which the listen socket of the new bus
                  server should be bound.
     @type port: int
+    @param tcpnodelay: If True, the socket will be set to TCP_NODELAY.
+    @type tcpnodelay: bool
     @param connector: A connector that should be attached to the bus
                       server.
     """
-    key = (host, port)
+    key = (host, port, tcpnodelay)
     with __busServersLock:
         bus = __busServers.get(key)
         if bus is None:
-            bus = BusServer(host, port)
+            bus = BusServer(host, port, tcpnodelay)
             bus.activate()
             __busServers[key] = bus
             bus.addConnector(connector)
@@ -510,7 +558,7 @@ class BusServer (Bus):
     @author: jmoringe
     """
 
-    def __init__(self, host, port, backlog = 5):
+    def __init__(self, host, port, tcpnodelay, backlog = 5):
         """
         @param host: A hostname or address identifying the interface
                      to which the listen socket of the new bus server
@@ -519,6 +567,8 @@ class BusServer (Bus):
         @param port: The port to which the listen socket of the new
                      bus server should be bound.
         @type port: int
+        @param tcpnodelay: If True, the socket will be set to TCP_NODELAY.
+        @type tcpnodelay: bool
         @param backlog: The maximum number of queued connection
                         attempts.
         @type backlog: int
@@ -529,6 +579,7 @@ class BusServer (Bus):
 
         self.__host           = host
         self.__port           = port
+        self.__tcpnodelay     = tcpnodelay
         self.__backlog        = backlog
         self.__socket         = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
         self.__acceptorThread = None
@@ -550,7 +601,7 @@ class BusServer (Bus):
                 if sys.platform == 'darwin':
                     clientSocket.settimeout(None)
                 self.__logger.info('Accepted client %s', addr)
-                self.addConnection(BusConnection(socket_ = clientSocket, isServer = True))
+                self.addConnection(BusConnection(socket_ = clientSocket, isServer = True, tcpnodelay = self.__tcpnodelay))
             except socket.timeout, e:
                 if sys.platform != 'darwin':
                     self.__logger.error('Unexpected timeout in acceptClients: "%s"', e)
@@ -646,9 +697,10 @@ class Connector(rsb.transport.Connector,
 
         self.__active = False
 
-        self.__bus    = None
-        self.__host   = options.get('host', 'localhost')
-        self.__port   = int(options.get('port', '55555'))
+        self.__bus        = None
+        self.__host       = options.get('host', 'localhost')
+        self.__port       = int(options.get('port', '55555'))
+        self.__tcpnodelay = options.get('nodelay', '1') in [ '1', 'true' ]
         serverString = options.get('server', 'auto')
         if serverString in [ '1', 'true' ]:
             self.__server = True
@@ -663,25 +715,25 @@ class Connector(rsb.transport.Connector,
         if self.__active:
             self.deactivate()
 
-    def __getBus(self, host, port, server):
+    def __getBus(self, host, port, tcpnodelay, server):
         self.__logger.info('Requested server role: %s', server)
 
         if server == True:
             self.__logger.info('Getting bus server %s:%d', host, port)
-            self.__bus = getBusServerFor(host, port, self)
+            self.__bus = getBusServerFor(host, port, tcpnodelay, self)
         elif server == False:
             self.__logger.info('Getting bus client %s:%d', host, port)
-            self.__bus = getBusClientFor(host, port, self)
+            self.__bus = getBusClientFor(host, port, tcpnodelay, self)
         elif server == 'auto':
             try:
                 self.__logger.info('Trying to get bus server %s:%d (in server = auto mode)',
                                    host, port)
-                self.__bus = getBusServerFor(host, port, self)
+                self.__bus = getBusServerFor(host, port, tcpnodelay, self)
             except Exception, e:
                 self.__logger.info('Failed to get bus server: %s', e)
                 self.__logger.info('Trying to get bus client %s:%d (in server = auto mode)',
                                    host, port)
-                self.__bus = getBusClientFor(host, port, self)
+                self.__bus = getBusClientFor(host, port, tcpnodelay, self)
         else:
             raise TypeError, 'server argument has to be True, False or "auto", not "%s"' % server
         self.__logger.info('Got %s', self.__bus)
@@ -698,7 +750,7 @@ class Connector(rsb.transport.Connector,
 
         self.__logger.info('Activating')
 
-        self.__bus = self.__getBus(self.__host, self.__port, self.__server)
+        self.__bus = self.__getBus(self.__host, self.__port, self.__tcpnodelay, self.__server)
 
         self.__active = True
 
